@@ -1,7 +1,7 @@
 /*
  * Copyright (C) 2010-2011 Robert Ancell.
  * Author: Robert Ancell <robert.ancell@canonical.com>
- * 
+ *
  * This program is free software: you can redistribute it and/or modify it under
  * the terms of the GNU General Public License as published by the Free Software
  * Foundation, either version 3 of the License, or (at your option) any later
@@ -19,6 +19,7 @@
 
 #include "greeter.h"
 #include "configuration.h"
+#include "shared-data-manager.h"
 
 enum {
     PROP_0,
@@ -43,11 +44,11 @@ struct GreeterPrivate
     guint8 *read_buffer;
     gsize n_read;
     gboolean use_secure_memory;
-  
+
     /* Hints for the greeter */
     GHashTable *hints;
 
-    /* Default session to use */   
+    /* Default session to use */
     gchar *default_session;
 
     /* Sequence number of current PAM session */
@@ -62,6 +63,9 @@ struct GreeterPrivate
     /* PAM session being constructed by the greeter */
     Session *authentication_session;
 
+    /* TRUE if a the greeter can handle a reset; else we will just kill it instead */
+    gboolean resettable;
+
     /* TRUE if a user has been authenticated and the session requested to start */
     gboolean start_session;
 
@@ -89,7 +93,8 @@ typedef enum
     GREETER_MESSAGE_START_SESSION,
     GREETER_MESSAGE_CANCEL_AUTHENTICATION,
     GREETER_MESSAGE_SET_LANGUAGE,
-    GREETER_MESSAGE_AUTHENTICATE_REMOTE
+    GREETER_MESSAGE_AUTHENTICATE_REMOTE,
+    GREETER_MESSAGE_ENSURE_SHARED_DIR,
 } GreeterMessage;
 
 /* Messages from the server to the greeter */
@@ -98,7 +103,10 @@ typedef enum
     SERVER_MESSAGE_CONNECTED = 0,
     SERVER_MESSAGE_PROMPT_AUTHENTICATION,
     SERVER_MESSAGE_END_AUTHENTICATION,
-    SERVER_MESSAGE_SESSION_RESULT
+    SERVER_MESSAGE_SESSION_RESULT,
+    SERVER_MESSAGE_SHARED_DIR_RESULT,
+    SERVER_MESSAGE_IDLE,
+    SERVER_MESSAGE_RESET,
 } ServerMessage;
 
 static gboolean read_cb (GIOChannel *source, GIOCondition condition, gpointer data);
@@ -124,6 +132,12 @@ greeter_set_allow_guest (Greeter *greeter, gboolean allow_guest)
     greeter->priv->allow_guest = allow_guest;
 }
 
+void
+greeter_clear_hints (Greeter *greeter)
+{
+    g_hash_table_remove_all (greeter->priv->hints);
+}
+
 void
 greeter_set_hint (Greeter *greeter, const gchar *name, const gchar *value)
 {
@@ -169,13 +183,31 @@ int_length (void)
 static void
 write_message (Greeter *greeter, guint8 *message, gsize message_length)
 {
+    gchar *data;
+    gsize data_length;
     GError *error = NULL;
 
-    g_io_channel_write_chars (greeter->priv->to_greeter_channel, (gchar *) message, message_length, NULL, &error);
+    data = (gchar *) message;
+    data_length = message_length;
+    while (data_length > 0)
+    {
+        GIOStatus status;
+        gsize n_written;
+
+        status = g_io_channel_write_chars (greeter->priv->to_greeter_channel, data, data_length, &n_written, &error);
+        if (error)
+            l_warning (greeter, "Error writing to greeter: %s", error->message);
+        g_clear_error (&error);
+        if (status != G_IO_STATUS_NORMAL)
+            return;
+        data_length -= n_written;
+        data += n_written;
+    }
+
+    g_io_channel_flush (greeter->priv->to_greeter_channel, &error);
     if (error)
-        l_warning (greeter, "Error writing to greeter: %s", error->message);
+        l_warning (greeter, "Failed to flush data to greeter: %s", error->message);
     g_clear_error (&error);
-    g_io_channel_flush (greeter->priv->to_greeter_channel, NULL);
 }
 
 static void
@@ -194,7 +226,7 @@ static void
 write_string (guint8 *buffer, gint buffer_length, const gchar *value, gsize *offset)
 {
     gint length;
-  
+
     if (value)
         length = strlen (value);
     else
@@ -226,7 +258,7 @@ string_length (const gchar *value)
 }
 
 static void
-handle_connect (Greeter *greeter, const gchar *version)
+handle_connect (Greeter *greeter, const gchar *version, gboolean resettable)
 {
     guint8 message[MAX_MESSAGE_LENGTH];
     gsize offset = 0;
@@ -234,7 +266,9 @@ handle_connect (Greeter *greeter, const gchar *version)
     GHashTableIter iter;
     gpointer key, value;
 
-    l_debug (greeter, "Greeter connected version=%s", version);
+    l_debug (greeter, "Greeter connected version=%s resettable=%s", version, resettable ? "true" : "false");
+
+    greeter->priv->resettable = resettable;
 
     length = string_length (VERSION);
     g_hash_table_iter_init (&iter, greeter->priv->hints);
@@ -273,7 +307,7 @@ pam_messages_cb (Session *session, Greeter *greeter)
     size = int_length () + string_length (session_get_username (session)) + int_length ();
     for (i = 0; i < messages_length; i++)
         size += int_length () + string_length (messages[i].msg);
-  
+
     write_header (message, MAX_MESSAGE_LENGTH, SERVER_MESSAGE_PROMPT_AUTHENTICATION, size, &offset);
     write_int (message, MAX_MESSAGE_LENGTH, greeter->priv->authentication_sequence_number, &offset);
     write_string (message, MAX_MESSAGE_LENGTH, session_get_username (session), &offset);
@@ -309,7 +343,40 @@ send_end_authentication (Greeter *greeter, guint32 sequence_number, const gchar
     write_int (message, MAX_MESSAGE_LENGTH, sequence_number, &offset);
     write_string (message, MAX_MESSAGE_LENGTH, username, &offset);
     write_int (message, MAX_MESSAGE_LENGTH, result, &offset);
-    write_message (greeter, message, offset); 
+    write_message (greeter, message, offset);
+}
+
+void
+greeter_idle (Greeter *greeter)
+{
+    guint8 message[MAX_MESSAGE_LENGTH];
+    gsize offset = 0;
+
+    write_header (message, MAX_MESSAGE_LENGTH, SERVER_MESSAGE_IDLE, 0, &offset);
+    write_message (greeter, message, offset);
+}
+
+void
+greeter_reset (Greeter *greeter)
+{
+    guint8 message[MAX_MESSAGE_LENGTH];
+    gsize offset = 0;
+    guint32 length = 0;
+    GHashTableIter iter;
+    gpointer key, value;
+
+    g_hash_table_iter_init (&iter, greeter->priv->hints);
+    while (g_hash_table_iter_next (&iter, &key, &value))
+        length += string_length (key) + string_length (value);
+
+    write_header (message, MAX_MESSAGE_LENGTH, SERVER_MESSAGE_RESET, length, &offset);
+    g_hash_table_iter_init (&iter, greeter->priv->hints);
+    while (g_hash_table_iter_next (&iter, &key, &value))
+    {
+        write_string (message, MAX_MESSAGE_LENGTH, key, &offset);
+        write_string (message, MAX_MESSAGE_LENGTH, value, &offset);
+    }
+    write_message (greeter, message, offset);
 }
 
 static void
@@ -417,7 +484,7 @@ handle_login_as_guest (Greeter *greeter, guint32 sequence_number)
         return;
     }
 
-    greeter->priv->guest_account_authenticated = TRUE;  
+    greeter->priv->guest_account_authenticated = TRUE;
     send_end_authentication (greeter, sequence_number, "", PAM_SUCCESS);
 }
 
@@ -627,6 +694,24 @@ handle_set_language (Greeter *greeter, const gchar *language)
     user_set_language (user, language);
 }
 
+static void
+handle_ensure_shared_dir (Greeter *greeter, const gchar *username)
+{
+    gchar *dir;
+    guint8 message[MAX_MESSAGE_LENGTH];
+    gsize offset = 0;
+
+    l_debug (greeter, "Greeter requests data directory for user %s", username);
+
+    dir = shared_data_manager_ensure_user_dir (shared_data_manager_get_instance (), username);
+
+    write_header (message, MAX_MESSAGE_LENGTH, SERVER_MESSAGE_SHARED_DIR_RESULT, string_length (dir), &offset);
+    write_string (message, MAX_MESSAGE_LENGTH, dir, &offset);
+    write_message (greeter, message, offset);
+
+    g_free (dir);
+}
+
 static guint32
 read_int (Greeter *greeter, gsize *offset)
 {
@@ -703,24 +788,29 @@ read_cb (GIOChannel *source, GIOCondition condition, gpointer data)
     Greeter *greeter = data;
     gsize n_to_read, n_read, offset;
     GIOStatus status;
-    int id, i;
+    int id, length, i;
     guint32 sequence_number, n_secrets, max_secrets;
     gchar *version, *username, *session_name, *language;
     gchar **secrets;
+    gboolean resettable = FALSE;
     GError *error = NULL;
 
     if (condition == G_IO_HUP)
     {
         l_debug (greeter, "Greeter closed communication channel");
+        greeter->priv->from_greeter_watch = 0;
         return FALSE;
     }
-  
+
     n_to_read = HEADER_SIZE;
     if (greeter->priv->n_read >= HEADER_SIZE)
     {
         n_to_read = get_message_length (greeter);
         if (n_to_read <= HEADER_SIZE)
+        {
+            greeter->priv->from_greeter_watch = 0;
             return FALSE;
+        }
     }
 
     status = g_io_channel_read_chars (greeter->priv->from_greeter_channel,
@@ -747,17 +837,19 @@ read_cb (GIOChannel *source, GIOCondition condition, gpointer data)
             greeter->priv->read_buffer = secure_realloc (greeter, greeter->priv->read_buffer, n_to_read);
             read_cb (source, condition, greeter);
             return TRUE;
-        }      
+        }
     }
-  
+
     offset = 0;
     id = read_int (greeter, &offset);
-    read_int (greeter, &offset);
+    length = HEADER_SIZE + read_int (greeter, &offset);
     switch (id)
     {
     case GREETER_MESSAGE_CONNECT:
         version = read_string (greeter, &offset);
-        handle_connect (greeter, version);
+        if (offset < length)
+            resettable = read_int (greeter, &offset) != 0;
+        handle_connect (greeter, version, resettable);
         g_free (version);
         break;
     case GREETER_MESSAGE_AUTHENTICATE:
@@ -782,6 +874,7 @@ read_cb (GIOChannel *source, GIOCondition condition, gpointer data)
         if (n_secrets > max_secrets)
         {
             l_warning (greeter, "Array length of %u elements too long", n_secrets);
+            greeter->priv->from_greeter_watch = 0;
             return FALSE;
         }
         secrets = g_malloc (sizeof (gchar *) * (n_secrets + 1));
@@ -806,6 +899,11 @@ read_cb (GIOChannel *source, GIOCondition condition, gpointer data)
         handle_set_language (greeter, language);
         g_free (language);
         break;
+    case GREETER_MESSAGE_ENSURE_SHARED_DIR:
+        username = read_string (greeter, &offset);
+        handle_ensure_shared_dir (greeter, username);
+        g_free (username);
+        break;
     default:
         l_warning (greeter, "Unknown message from greeter: %d", id);
         break;
@@ -830,6 +928,13 @@ greeter_get_authentication_session (Greeter *greeter)
     return greeter->priv->authentication_session;
 }
 
+gboolean
+greeter_get_resettable (Greeter *greeter)
+{
+    g_return_val_if_fail (greeter != NULL, FALSE);
+    return greeter->priv->resettable;
+}
+
 gboolean
 greeter_get_start_session (Greeter *greeter)
 {
@@ -851,6 +956,7 @@ greeter_start (Session *session)
     int to_greeter_pipe[2], from_greeter_pipe[2];
     gboolean result = FALSE;
     gchar *value;
+    GError *error = NULL;
 
     /* Create a pipe to talk with the greeter */
     if (pipe (to_greeter_pipe) != 0 || pipe (from_greeter_pipe) != 0)
@@ -859,9 +965,15 @@ greeter_start (Session *session)
         return FALSE;
     }
     greeter->priv->to_greeter_channel = g_io_channel_unix_new (to_greeter_pipe[1]);
-    g_io_channel_set_encoding (greeter->priv->to_greeter_channel, NULL, NULL);
+    g_io_channel_set_encoding (greeter->priv->to_greeter_channel, NULL, &error);
+    if (error)
+        g_warning ("Failed to set encoding on to greeter channel to binary: %s\n", error->message);
+    g_clear_error (&error);
     greeter->priv->from_greeter_channel = g_io_channel_unix_new (from_greeter_pipe[0]);
-    g_io_channel_set_encoding (greeter->priv->from_greeter_channel, NULL, NULL);
+    g_io_channel_set_encoding (greeter->priv->from_greeter_channel, NULL, &error);
+    if (error)
+        g_warning ("Failed to set encoding on from greeter channel to binary: %s\n", error->message);
+    g_clear_error (&error);
     g_io_channel_set_buffered (greeter->priv->from_greeter_channel, FALSE);
     greeter->priv->from_greeter_watch = g_io_add_watch (greeter->priv->from_greeter_channel, G_IO_IN | G_IO_HUP, read_cb, greeter);
 
