@@ -1,7 +1,7 @@
 /*
  * Copyright (C) 2010-2011 Robert Ancell.
  * Author: Robert Ancell <robert.ancell@canonical.com>
- * 
+ *
  * This program is free software: you can redistribute it and/or modify it under
  * the terms of the GNU General Public License as published by the Free Software
  * Foundation, either version 3 of the License, or (at your option) any later
@@ -30,11 +30,14 @@ static guint signals[LAST_SIGNAL] = { 0 };
 
 struct SeatPrivate
 {
+    /* XDG name for this seat */
+    gchar *name;
+
     /* Configuration for this seat */
     GHashTable *properties;
 
-    /* TRUE if able to switch users */
-    gboolean can_switch;
+    /* TRUE if this seat can run multiple sessions at once */
+    gboolean supports_multi_session;
 
     /* TRUE if display server can be shared for sessions */
     gboolean share_display_server;
@@ -53,7 +56,7 @@ struct SeatPrivate
 
     /* The session to set active when it starts */
     Session *session_to_activate;
-  
+
     /* TRUE once we have started */
     gboolean started;
 
@@ -64,11 +67,6 @@ struct SeatPrivate
     gboolean stopped;
 };
 
-/* PAM services to use */
-#define GREETER_SERVICE   "lightdm-greeter"
-#define USER_SERVICE      "lightdm"
-#define AUTOLOGIN_SERVICE "lightdm-autologin"
-
 static void seat_logger_iface_init (LoggerInterface *iface);
 
 G_DEFINE_TYPE_WITH_CODE (Seat, seat, G_TYPE_OBJECT,
@@ -83,7 +81,7 @@ typedef struct
 static GHashTable *seat_modules = NULL;
 
 // FIXME: Make a get_display_server() that re-uses display servers if supported
-static DisplayServer *create_display_server (Seat *seat, const gchar *session_type);
+static DisplayServer *create_display_server (Seat *seat, Session *session);
 static Greeter *create_greeter_session (Seat *seat);
 static void start_session (Seat *seat, Session *session);
 
@@ -104,11 +102,11 @@ seat_register_module (const gchar *name, GType type)
 }
 
 Seat *
-seat_new (const gchar *module_name)
+seat_new (const gchar *module_name, const gchar *name)
 {
     Seat *seat;
     SeatModule *m = NULL;
-  
+
     g_return_val_if_fail (module_name != NULL, NULL);
 
     if (seat_modules)
@@ -117,6 +115,7 @@ seat_new (const gchar *module_name)
         return NULL;
 
     seat = g_object_new (m->type, NULL);
+    seat->priv->name = g_strdup (name);
 
     return seat;
 }
@@ -135,6 +134,13 @@ seat_get_string_property (Seat *seat, const gchar *name)
     return g_hash_table_lookup (seat->priv->properties, name);
 }
 
+gchar **
+seat_get_string_list_property (Seat *seat, const gchar *name)
+{
+    g_return_val_if_fail (seat != NULL, NULL);
+    return g_strsplit (g_hash_table_lookup (seat->priv->properties, name), ";", 0);
+}
+
 gboolean
 seat_get_boolean_property (Seat *seat, const gchar *name)
 {
@@ -150,19 +156,23 @@ seat_get_integer_property (Seat *seat, const gchar *name)
     return value ? atoi (value) : 0;
 }
 
+const gchar *
+seat_get_name (Seat *seat)
+{
+    return seat->priv->name;
+}
+
 void
-seat_set_can_switch (Seat *seat, gboolean can_switch)
+seat_set_supports_multi_session (Seat *seat, gboolean supports_multi_session)
 {
     g_return_if_fail (seat != NULL);
-
-    seat->priv->can_switch = can_switch;
+    seat->priv->supports_multi_session = supports_multi_session;
 }
 
 void
 seat_set_share_display_server (Seat *seat, gboolean share_display_server)
 {
     g_return_if_fail (seat != NULL);
-
     seat->priv->share_display_server = share_display_server;
 }
 
@@ -170,9 +180,13 @@ gboolean
 seat_start (Seat *seat)
 {
     g_return_val_if_fail (seat != NULL, FALSE);
-  
+
+    l_debug (seat, "Starting");
+
     SEAT_GET_CLASS (seat)->setup (seat);
-    return SEAT_GET_CLASS (seat)->start (seat);
+    seat->priv->started = SEAT_GET_CLASS (seat)->start (seat);
+
+    return seat->priv->started;
 }
 
 GList *
@@ -182,6 +196,13 @@ seat_get_sessions (Seat *seat)
     return seat->priv->sessions;
 }
 
+static gboolean
+set_greeter_idle (gpointer greeter)
+{
+    greeter_idle (GREETER (greeter));
+    return FALSE;
+}
+
 void
 seat_set_active_session (Seat *seat, Session *session)
 {
@@ -201,8 +222,24 @@ seat_set_active_session (Seat *seat, Session *session)
 
         if (IS_GREETER (s))
         {
-            l_debug (seat, "Stopping greeter");
-            session_stop (s);
+            Greeter *greeter = GREETER (s);
+            if (greeter_get_resettable (greeter))
+            {
+                if (seat->priv->active_session == SESSION (greeter))
+                {
+                    l_debug (seat, "Idling greeter");
+                    /* Do this in an idle callback, because we might very well
+                       be in the middle of responding to a START_SESSION
+                       request by a greeter.  So they won't expect an IDLE
+                       call during that.  Plus, this isn't time-sensitive. */
+                    g_idle_add (set_greeter_idle, greeter);
+                }
+            }
+            else
+            {
+                l_debug (seat, "Stopping greeter");
+                session_stop (s);
+            }
         }
     }
 
@@ -213,6 +250,8 @@ seat_set_active_session (Seat *seat, Session *session)
             session_lock (seat->priv->active_session);
         g_object_unref (seat->priv->active_session);
     }
+
+    session_activate (session);
     seat->priv->active_session = g_object_ref (session);
 }
 
@@ -234,7 +273,7 @@ gboolean
 seat_get_can_switch (Seat *seat)
 {
     g_return_val_if_fail (seat != NULL, FALSE);
-    return seat->priv->can_switch;
+    return seat_get_boolean_property (seat, "allow-user-switching") && seat->priv->supports_multi_session;
 }
 
 gboolean
@@ -249,8 +288,8 @@ run_script (Seat *seat, DisplayServer *display_server, const gchar *script_name,
 {
     Process *script;
     gboolean result = FALSE;
-  
-    script = process_new ();
+
+    script = process_new (NULL, NULL);
 
     process_set_command (script, script_name);
 
@@ -299,7 +338,7 @@ run_script (Seat *seat, DisplayServer *display_server, const gchar *script_name,
 
 static void
 seat_real_run_script (Seat *seat, DisplayServer *display_server, Process *process)
-{  
+{
 }
 
 static void
@@ -330,20 +369,20 @@ check_stopped (Seat *seat)
     }
 }
 
-static gboolean
-get_start_local_sessions (Seat *seat)
-{
-    return SEAT_GET_CLASS (seat)->get_start_local_sessions (seat);
-}
-
 static void
 display_server_stopped_cb (DisplayServer *display_server, Seat *seat)
 {
+    const gchar *script;
     GList *list, *link;
     Session *active_session;
 
     l_debug (seat, "Display server stopped");
 
+    /* Run a script right after stopping the display server */
+    script = seat_get_string_property (seat, "display-stopped-script");
+    if (script)
+        run_script (seat, NULL, script, NULL);
+
     g_signal_handlers_disconnect_matched (display_server, G_SIGNAL_MATCH_DATA, 0, 0, NULL, NULL, seat);
     seat->priv->display_servers = g_list_remove (seat->priv->display_servers, display_server);
 
@@ -379,14 +418,18 @@ display_server_stopped_cb (DisplayServer *display_server, Seat *seat)
     }
     g_list_free_full (list, g_object_unref);
 
-    if (!seat->priv->stopping && get_start_local_sessions (seat))
+    if (!seat->priv->stopping)
     {
         /* If we were the active session, switch to a greeter */
         active_session = seat_get_active_session (seat);
         if (!active_session || session_get_display_server (active_session) == display_server)
         {
             l_debug (seat, "Active display server stopped, starting greeter");
-            seat_switch_to_greeter (seat);
+            if (!seat_switch_to_greeter (seat))
+            {
+                l_debug (seat, "Stopping; failed to start a greeter");
+                seat_stop (seat);
+            }
         }
     }
 
@@ -399,38 +442,101 @@ can_share_display_server (Seat *seat, DisplayServer *display_server)
     return seat->priv->share_display_server && display_server_get_can_share (display_server);
 }
 
+static Greeter *
+find_greeter_session (Seat *seat)
+{
+    GList *link;
+
+    for (link = seat->priv->sessions; link; link = link->next)
+    {
+        Session *session = link->data;
+        if (!session_get_is_stopping (session) && IS_GREETER (session))
+            return GREETER (session);
+    }
+
+    return NULL;
+}
+
+static Greeter *
+find_resettable_greeter (Seat *seat)
+{
+    GList *link;
+
+    for (link = seat->priv->sessions; link; link = link->next)
+    {
+        Session *session = link->data;
+        if (!session_get_is_stopping (session) && IS_GREETER (session) &&
+            greeter_get_resettable (GREETER (session)))
+            return GREETER (session);
+    }
+
+    return NULL;
+}
+
+static void
+set_greeter_hints (Seat *seat, Greeter *greeter_session)
+{
+    greeter_clear_hints (greeter_session);
+    greeter_set_hint (greeter_session, "default-session", seat_get_string_property (seat, "user-session"));
+    greeter_set_hint (greeter_session, "hide-users", seat_get_boolean_property (seat, "greeter-hide-users") ? "true" : "false");
+    greeter_set_hint (greeter_session, "show-manual-login", seat_get_boolean_property (seat, "greeter-show-manual-login") ? "true" : "false");
+    greeter_set_hint (greeter_session, "show-remote-login", seat_get_boolean_property (seat, "greeter-show-remote-login") ? "true" : "false");
+    greeter_set_hint (greeter_session, "has-guest-account", seat_get_allow_guest (seat) && seat_get_boolean_property (seat, "greeter-allow-guest") ? "true" : "false");
+}
+
 static void
 switch_to_greeter_from_failed_session (Seat *seat, Session *session)
 {
     Greeter *greeter_session;
+    gboolean existing = FALSE;
+
+    /* Switch to greeter if one open */
+    greeter_session = find_resettable_greeter (seat);
+    if (greeter_session)
+    {
+        l_debug (seat, "Switching to existing greeter");
+        set_greeter_hints (seat, greeter_session);
+        existing = TRUE;
+    }
+    else
+    {
+        greeter_session = create_greeter_session (seat);
+    }
 
-    greeter_session = create_greeter_session (seat);
     if (session_get_is_guest (session))
         greeter_set_hint (greeter_session, "select-guest", "true");
     else
         greeter_set_hint (greeter_session, "select-user", session_get_username (session));
-    if (seat->priv->session_to_activate)
-        g_object_unref (seat->priv->session_to_activate);
-    seat->priv->session_to_activate = g_object_ref (greeter_session);
 
-    if (can_share_display_server (seat, session_get_display_server (session)))
-        session_set_display_server (SESSION (greeter_session), session_get_display_server (session));
+    if (existing)
+    {
+        greeter_reset (greeter_session);
+        seat_set_active_session (seat, SESSION (greeter_session));
+    }
     else
     {
-        DisplayServer *display_server;
+        if (seat->priv->session_to_activate)
+            g_object_unref (seat->priv->session_to_activate);
+        seat->priv->session_to_activate = g_object_ref (greeter_session);
 
-        display_server = create_display_server (seat, session_get_session_type (session));
-        if (!display_server_start (display_server))
+        if (can_share_display_server (seat, session_get_display_server (session)))
+            session_set_display_server (SESSION (greeter_session), session_get_display_server (session));
+        else
         {
-            l_debug (seat, "Failed to start display server for greeter");
-            seat_stop (seat);
+            DisplayServer *display_server;
+
+            display_server = create_display_server (seat, session);
+            session_set_display_server (session, display_server);
+            if (!display_server_start (display_server))
+            {
+                l_debug (seat, "Failed to start display server for greeter");
+                seat_stop (seat);
+            }
         }
 
-        session_set_display_server (session, display_server);
+        start_session (seat, SESSION (greeter_session));
     }
 
-    start_session (seat, SESSION (greeter_session));
-
     /* Stop failed session */
     session_stop (session);
 }
@@ -498,6 +604,17 @@ run_session (Seat *seat, Session *session)
         g_object_unref (seat->priv->session_to_activate);
         seat->priv->session_to_activate = NULL;
     }
+    else if (seat->priv->active_session)
+    {
+        /* Multiple sessions can theoretically be on the same VT (especially
+           if using Mir).  If a new session appears on an existing active VT,
+           logind will mark it as active, while ConsoleKit will re-mark the
+           oldest session as active.  In either case, that may not be the
+           session that we want to be active.  So let's be explicit and
+           re-activate the correct session whenever a new session starts.
+           There's no harm to do this in seats that enforce separate VTs. */
+        session_activate (seat->priv->active_session);
+    }
 }
 
 static Session *
@@ -665,7 +782,11 @@ session_stopped_cb (Session *session, Seat *seat)
     else if (!IS_GREETER (session) && session == seat_get_active_session (seat))
     {
         l_debug (seat, "Active session stopped, starting greeter");
-        seat_switch_to_greeter (seat);
+        if (!seat_switch_to_greeter (seat))
+        {
+            l_debug (seat, "Stopping; failed to start a greeter");
+            seat_stop (seat);
+        }
     }
 
     /* Stop the display server if no-longer required */
@@ -772,7 +893,7 @@ get_session_argv (Seat *seat, SessionConfig *session_config, const gchar *sessio
         g_free (argv[0]);
         argv[0] = path;
     }
-  
+
     return argv;
 }
 
@@ -814,7 +935,8 @@ configure_session (Session *session, SessionConfig *config, const gchar *session
 {
     const gchar *desktop_name;
 
-    session_set_session_type (session, session_config_get_session_type (config));
+    session_set_config (session, config);
+    session_set_env (session, "XDG_SESSION_DESKTOP", session_name);
     session_set_env (session, "DESKTOP_SESSION", session_name);
     session_set_env (session, "GDMSESSION", session_name);
     desktop_name = session_config_get_desktop_name (config);
@@ -867,7 +989,7 @@ create_user_session (Seat *seat, const gchar *username, gboolean autostart)
         g_object_unref (session_config);
     }
     else
-        l_debug (seat, "Can't find session '%s'", seat_get_string_property (seat, "user-session"));
+        l_debug (seat, "Can't find session '%s'", session_name);
 
     g_object_unref (user);
 
@@ -892,20 +1014,23 @@ prepend_argv (gchar ***argv, const gchar *value)
 }
 
 static Session *
-create_guest_session (Seat *seat)
+create_guest_session (Seat *seat, const gchar *session_name)
 {
-    const gchar *session_name, *guest_wrapper;
+    const gchar *guest_wrapper;
     gchar *sessions_dir, **argv;
     SessionConfig *session_config;
     Session *session;
 
-    session_name = seat_get_string_property (seat, "user-session");
+    if (!session_name)
+        session_name = seat_get_string_property (seat, "guest-session");
+    if (!session_name)
+        session_name = seat_get_string_property (seat, "user-session");
     sessions_dir = config_get_string (config_get_instance (), "LightDM", "sessions-directory");
     session_config = find_session_config (seat, sessions_dir, session_name);
     g_free (sessions_dir);
     if (!session_config)
     {
-        l_debug (seat, "Can't find session '%s'", seat_get_string_property (seat, "user-session"));
+        l_debug (seat, "Can't find session '%s'", session_name);
         return NULL;
     }
 
@@ -926,7 +1051,7 @@ create_guest_session (Seat *seat)
     session_set_argv (session, argv);
     g_strfreev (argv);
     g_object_unref (session_config);
-  
+
     return session;
 }
 
@@ -936,7 +1061,7 @@ greeter_create_session_cb (Greeter *greeter, Seat *seat)
     Session *session;
 
     session = create_session (seat, FALSE);
-    session_set_session_type (session, session_get_session_type (SESSION (greeter)));
+    session_set_config (session, session_get_config (SESSION (greeter)));
     session_set_display_server (session, session_get_display_server (SESSION (greeter)));
 
     return g_object_ref (session);
@@ -952,10 +1077,10 @@ greeter_start_session_cb (Greeter *greeter, SessionType type, const gchar *sessi
     /* Get the session to use */
     if (greeter_get_guest_authenticated (greeter))
     {
-        session = create_guest_session (seat);
+        session = create_guest_session (seat, session_name);
         if (!session)
             return FALSE;
-        session_set_pam_service (session, AUTOLOGIN_SERVICE);
+        session_set_pam_service (session, seat_get_string_property (seat, "pam-autologin-service"));
     }
     else
     {
@@ -966,7 +1091,7 @@ greeter_start_session_cb (Greeter *greeter, SessionType type, const gchar *sessi
         gchar **argv;
 
         session = greeter_get_authentication_session (greeter);
-  
+
         /* Get session command to run */
         switch (type)
         {
@@ -996,7 +1121,7 @@ greeter_start_session_cb (Greeter *greeter, SessionType type, const gchar *sessi
         g_free (sessions_dir);
         if (!session_config)
         {
-            l_debug (seat, "Can't find session '%s'", seat_get_string_property (seat, "user-session"));
+            l_debug (seat, "Can't find session '%s'", session_name);
             return FALSE;
         }
 
@@ -1026,7 +1151,8 @@ greeter_start_session_cb (Greeter *greeter, SessionType type, const gchar *sessi
 
     /* If can re-use the display server, stop the greeter first */
     display_server = session_get_display_server (SESSION (greeter));
-    if (can_share_display_server (seat, display_server) &&
+    if (!greeter_get_resettable (greeter) &&
+        can_share_display_server (seat, display_server) &&
         strcmp (display_server_get_session_type (display_server), session_get_session_type (session)) == 0)
     {
         l_debug (seat, "Stopping greeter; display server will be re-used for user session");
@@ -1042,7 +1168,7 @@ greeter_start_session_cb (Greeter *greeter, SessionType type, const gchar *sessi
     /* Otherwise start a new display server for this session */
     else
     {
-        display_server = create_display_server (seat, session_get_session_type (session));
+        display_server = create_display_server (seat, session);
         session_set_display_server (session, display_server);
         if (!display_server_start (display_server))
         {
@@ -1061,6 +1187,9 @@ create_greeter_session (Seat *seat)
     SessionConfig *session_config;
     Greeter *greeter_session;
     const gchar *greeter_wrapper;
+    const gchar *autologin_username;
+    int autologin_timeout;
+    gboolean autologin_guest;
 
     l_debug (seat, "Creating greeter session");
 
@@ -1081,19 +1210,19 @@ create_greeter_session (Seat *seat)
     }
 
     greeter_session = SEAT_GET_CLASS (seat)->create_greeter_session (seat);
-    session_set_session_type (SESSION (greeter_session), session_config_get_session_type (session_config));
+    session_set_config (SESSION (greeter_session), session_config);
     seat->priv->sessions = g_list_append (seat->priv->sessions, SESSION (greeter_session));
     g_signal_connect (greeter_session, "notify::active-username", G_CALLBACK (greeter_active_username_changed_cb), seat);
     g_signal_connect (greeter_session, "authentication-complete", G_CALLBACK (session_authentication_complete_cb), seat);
     g_signal_connect (greeter_session, "stopped", G_CALLBACK (session_stopped_cb), seat);
-  
+
     set_session_env (SESSION (greeter_session));
     session_set_env (SESSION (greeter_session), "XDG_SESSION_CLASS", "greeter");
 
-    session_set_pam_service (SESSION (greeter_session), GREETER_SERVICE);
+    session_set_pam_service (SESSION (greeter_session), seat_get_string_property (seat, "pam-greeter-service"));
     if (getuid () == 0)
     {
-        gchar *greeter_user;      
+        gchar *greeter_user;
         greeter_user = config_get_string (config_get_instance (), "LightDM", "greeter-user");
         session_set_username (SESSION (greeter_session), greeter_user);
         g_free (greeter_user);
@@ -1106,17 +1235,34 @@ create_greeter_session (Seat *seat)
     session_set_argv (SESSION (greeter_session), argv);
     g_strfreev (argv);
 
-    greeter_set_pam_services (greeter_session, USER_SERVICE, AUTOLOGIN_SERVICE);
+    greeter_set_pam_services (greeter_session,
+                              seat_get_string_property (seat, "pam-service"),
+                              seat_get_string_property (seat, "pam-autologin-service"));
     g_signal_connect (greeter_session, "create-session", G_CALLBACK (greeter_create_session_cb), seat);
     g_signal_connect (greeter_session, "start-session", G_CALLBACK (greeter_start_session_cb), seat);
 
     /* Set hints to greeter */
-    greeter_set_hint (greeter_session, "default-session", seat_get_string_property (seat, "user-session"));
     greeter_set_allow_guest (greeter_session, seat_get_allow_guest (seat));
-    greeter_set_hint (greeter_session, "hide-users", seat_get_boolean_property (seat, "greeter-hide-users") ? "true" : "false");
-    greeter_set_hint (greeter_session, "show-manual-login", seat_get_boolean_property (seat, "greeter-show-manual-login") ? "true" : "false");
-    greeter_set_hint (greeter_session, "show-remote-login", seat_get_boolean_property (seat, "greeter-show-remote-login") ? "true" : "false");
-    greeter_set_hint (greeter_session, "has-guest-account", seat_get_allow_guest (seat) && seat_get_boolean_property (seat, "greeter-allow-guest") ? "true" : "false");
+    set_greeter_hints (seat, greeter_session);
+
+    /* Configure for automatic login */
+    autologin_username = seat_get_string_property (seat, "autologin-user");
+    if (g_strcmp0 (autologin_username, "") == 0)
+        autologin_username = NULL;
+    autologin_timeout = seat_get_integer_property (seat, "autologin-user-timeout");
+    autologin_guest = seat_get_boolean_property (seat, "autologin-guest");
+    if (autologin_timeout > 0)
+    {
+        gchar *value;
+
+        value = g_strdup_printf ("%d", autologin_timeout);
+        greeter_set_hint (greeter_session, "autologin-timeout", value);
+        g_free (value);
+        if (autologin_username)
+            greeter_set_hint (greeter_session, "autologin-user", autologin_username);
+        if (autologin_guest)
+            greeter_set_hint (greeter_session, "autologin-guest", "true");
+    }
 
     g_object_unref (session_config);
 
@@ -1153,10 +1299,6 @@ display_server_ready_cb (DisplayServer *display_server, Seat *seat)
         return;
     }
 
-    /* Stop if don't need to run a session */
-    if (!get_start_local_sessions (seat))
-        return;
-
     emit_upstart_signal ("login-session-start");
 
     /* Start the session waiting for this display server */
@@ -1182,13 +1324,13 @@ display_server_ready_cb (DisplayServer *display_server, Seat *seat)
 }
 
 static DisplayServer *
-create_display_server (Seat *seat, const gchar *session_type)
+create_display_server (Seat *seat, Session *session)
 {
     DisplayServer *display_server;
 
-    l_debug (seat, "Creating display server of type %s", session_type);
+    l_debug (seat, "Creating display server of type %s", session_get_session_type (session));
 
-    display_server = SEAT_GET_CLASS (seat)->create_display_server (seat, session_type);
+    display_server = SEAT_GET_CLASS (seat)->create_display_server (seat, session);
     if (!display_server)
         return NULL;
 
@@ -1199,21 +1341,6 @@ create_display_server (Seat *seat, const gchar *session_type)
     return display_server;
 }
 
-static Greeter *
-find_greeter_session (Seat *seat)
-{
-    GList *link;
-
-    for (link = seat->priv->sessions; link; link = link->next)
-    {
-        Session *session = link->data;
-        if (!session_get_is_stopping (session) && IS_GREETER (session))
-            return GREETER (session);
-    }
-
-    return NULL;
-}
-
 gboolean
 seat_switch_to_greeter (Seat *seat)
 {
@@ -1222,10 +1349,10 @@ seat_switch_to_greeter (Seat *seat)
 
     g_return_val_if_fail (seat != NULL, FALSE);
 
-    if (!seat->priv->can_switch)
+    if (!seat_get_can_switch (seat))
         return FALSE;
 
-    /* Switch to greeter if one open (shouldn't be though) */
+    /* Switch to greeter if one open */
     greeter_session = find_greeter_session (seat);
     if (greeter_session)
     {
@@ -1235,16 +1362,17 @@ seat_switch_to_greeter (Seat *seat)
     }
 
     greeter_session = create_greeter_session (seat);
+    if (!greeter_session)
+        return FALSE;
+
     if (seat->priv->session_to_activate)
         g_object_unref (seat->priv->session_to_activate);
     seat->priv->session_to_activate = g_object_ref (greeter_session);
 
-    display_server = create_display_server (seat, session_get_session_type (SESSION (greeter_session)));
+    display_server = create_display_server (seat, SESSION (greeter_session));
     session_set_display_server (SESSION (greeter_session), display_server);
-    if (!display_server_start (display_server))
-        return FALSE;
 
-    return TRUE;
+    return display_server_start (display_server);
 }
 
 static void
@@ -1252,6 +1380,7 @@ switch_authentication_complete_cb (Session *session, Seat *seat)
 {
     Greeter *greeter_session;
     DisplayServer *display_server;
+    gboolean existing = FALSE;
 
     /* If authenticated, then unlock existing session or start new one */
     if (session_get_is_authenticated (session))
@@ -1272,7 +1401,7 @@ switch_authentication_complete_cb (Session *session, Seat *seat)
             if (seat->priv->session_to_activate)
                 g_object_unref (seat->priv->session_to_activate);
             seat->priv->session_to_activate = g_object_ref (session);
-            display_server = create_display_server (seat, session_get_session_type (session));
+            display_server = create_display_server (seat, session);
             session_set_display_server (session, display_server);
             display_server_start (display_server);
         }
@@ -1280,22 +1409,42 @@ switch_authentication_complete_cb (Session *session, Seat *seat)
         return;
     }
 
-    l_debug (seat, "Switching to greeter to authenticate session");
-
     session_stop (session);
 
-    greeter_session = create_greeter_session (seat);
+    /* See if we already have a greeter up and reuse it if so */
+    greeter_session = find_resettable_greeter (seat);
+    if (greeter_session)
+    {
+        l_debug (seat, "Switching to existing greeter to authenticate session");
+        set_greeter_hints (seat, greeter_session);
+        existing = TRUE;
+    }
+    else
+    {
+        l_debug (seat, "Starting greeter to authenticate session");
+        greeter_session = create_greeter_session (seat);
+    }
+
     if (session_get_is_guest (session))
         greeter_set_hint (greeter_session, "select-guest", "true");
     else
         greeter_set_hint (greeter_session, "select-user", session_get_username (session));
-    if (seat->priv->session_to_activate)
-        g_object_unref (seat->priv->session_to_activate);
-    seat->priv->session_to_activate = g_object_ref (greeter_session);
 
-    display_server = create_display_server (seat, session_get_session_type (SESSION (greeter_session)));
-    session_set_display_server (SESSION (greeter_session), display_server);
-    display_server_start (display_server);
+    if (existing)
+    {
+        greeter_reset (greeter_session);
+        seat_set_active_session (seat, SESSION (greeter_session));
+    }
+    else
+    {
+        if (seat->priv->session_to_activate)
+            g_object_unref (seat->priv->session_to_activate);
+        seat->priv->session_to_activate = g_object_ref (greeter_session);
+
+        display_server = create_display_server (seat, SESSION (greeter_session));
+        session_set_display_server (SESSION (greeter_session), display_server);
+        display_server_start (display_server);
+    }
 }
 
 gboolean
@@ -1306,7 +1455,7 @@ seat_switch_to_user (Seat *seat, const gchar *username, const gchar *session_nam
     g_return_val_if_fail (seat != NULL, FALSE);
     g_return_val_if_fail (username != NULL, FALSE);
 
-    if (!seat->priv->can_switch)
+    if (!seat_get_can_switch (seat))
         return FALSE;
 
     /* If we're already on this session, then ignore */
@@ -1319,10 +1468,9 @@ seat_switch_to_user (Seat *seat, const gchar *username, const gchar *session_nam
     /* Attempt to authenticate them */
     session = create_user_session (seat, username, FALSE);
     g_signal_connect (session, "authentication-complete", G_CALLBACK (switch_authentication_complete_cb), seat);
-    session_set_pam_service (session, USER_SERVICE);
-    session_start (session);
+    session_set_pam_service (session, seat_get_string_property (seat, "pam-service"));
 
-    return FALSE;
+    return session_start (session);
 }
 
 static Session *
@@ -1348,7 +1496,7 @@ seat_switch_to_guest (Seat *seat, const gchar *session_name)
 
     g_return_val_if_fail (seat != NULL, FALSE);
 
-    if (!seat->priv->can_switch || !seat_get_allow_guest (seat))
+    if (!seat_get_can_switch (seat) || !seat_get_allow_guest (seat))
         return FALSE;
 
     /* Switch to session if one open */
@@ -1360,21 +1508,19 @@ seat_switch_to_guest (Seat *seat, const gchar *session_name)
         return TRUE;
     }
 
-    session = create_guest_session (seat);
+    session = create_guest_session (seat, session_name);
     if (!session)
         return FALSE;
 
-    display_server = create_display_server (seat, session_get_session_type (session));
-    if (!display_server_start (display_server))
-        return FALSE;
+    display_server = create_display_server (seat, session);
 
     if (seat->priv->session_to_activate)
         g_object_unref (seat->priv->session_to_activate);
     seat->priv->session_to_activate = g_object_ref (session);
-    session_set_pam_service (session, AUTOLOGIN_SERVICE);
+    session_set_pam_service (session, seat_get_string_property (seat, "pam-autologin-service"));
     session_set_display_server (session, display_server);
 
-    return TRUE;
+    return display_server_start (display_server);
 }
 
 gboolean
@@ -1382,40 +1528,51 @@ seat_lock (Seat *seat, const gchar *username)
 {
     Greeter *greeter_session;
     DisplayServer *display_server;
+    gboolean existing = FALSE;
 
     g_return_val_if_fail (seat != NULL, FALSE);
 
-    if (!seat->priv->can_switch)
+    if (!seat_get_can_switch (seat))
         return FALSE;
 
     l_debug (seat, "Locking");
 
-    /* Switch to greeter if one open (shouldn't be though) */
-    greeter_session = find_greeter_session (seat);
+    /* Switch to greeter we can reuse */
+    greeter_session = find_resettable_greeter (seat);
     if (greeter_session)
     {
         l_debug (seat, "Switching to existing greeter");
-        seat_set_active_session (seat, SESSION (greeter_session));
-        return TRUE;
+        set_greeter_hints (seat, greeter_session);
+        existing = TRUE;
+    }
+    else
+    {
+        greeter_session = create_greeter_session (seat);
+        if (!greeter_session)
+            return FALSE;
     }
 
-    greeter_session = create_greeter_session (seat);
-    if (!greeter_session)
-        return FALSE;
-
-    display_server = create_display_server (seat, session_get_session_type (SESSION (greeter_session)));
-    if (!display_server_start (display_server))
-        return FALSE;
-
-    if (seat->priv->session_to_activate)
-        g_object_unref (seat->priv->session_to_activate);
-    seat->priv->session_to_activate = g_object_ref (greeter_session);
     greeter_set_hint (greeter_session, "lock-screen", "true");
     if (username)
         greeter_set_hint (greeter_session, "select-user", username);
-    session_set_display_server (SESSION (greeter_session), display_server);
 
-    return TRUE;
+    if (existing)
+    {
+        greeter_reset (greeter_session);
+        seat_set_active_session (seat, SESSION (greeter_session));
+        return TRUE;
+    }
+    else
+    {
+        display_server = create_display_server (seat, SESSION (greeter_session));
+
+        if (seat->priv->session_to_activate)
+            g_object_unref (seat->priv->session_to_activate);
+        seat->priv->session_to_activate = g_object_ref (greeter_session);
+        session_set_display_server (SESSION (greeter_session), display_server);
+
+        return display_server_start (display_server);
+    }
 }
 
 void
@@ -1438,12 +1595,6 @@ seat_get_is_stopping (Seat *seat)
     return seat->priv->stopping;
 }
 
-static gboolean
-seat_real_get_start_local_sessions (Seat *seat)
-{
-    return TRUE;
-}
-
 static void
 seat_real_setup (Seat *seat)
 {
@@ -1458,16 +1609,6 @@ seat_real_start (Seat *seat)
     gboolean autologin_in_background;
     Session *session = NULL, *background_session = NULL;
 
-    l_debug (seat, "Starting");
-
-    /* If this display server doesn't have a session running on it, just start it */
-    if (!get_start_local_sessions (seat))
-    {
-        DisplayServer *display_server;
-        display_server = create_display_server (seat, "x"); // FIXME: Not necessarily an X seat, but not sure what to put here
-        return display_server_start (display_server);
-    }
-
     /* Get autologin settings */
     autologin_username = seat_get_string_property (seat, "autologin-user");
     if (g_strcmp0 (autologin_username, "") == 0)
@@ -1480,12 +1621,12 @@ seat_real_start (Seat *seat)
     if (autologin_timeout == 0 || autologin_in_background)
     {
         if (autologin_guest)
-            session = create_guest_session (seat);
+            session = create_guest_session (seat, NULL);
         else if (autologin_username != NULL)
             session = create_user_session (seat, autologin_username, TRUE);
 
         if (session)
-            session_set_pam_service (session, AUTOLOGIN_SERVICE);
+            session_set_pam_service (session, seat_get_string_property (seat, "pam-autologin-service"));
 
         /* Load in background if required */
         if (autologin_in_background && session)
@@ -1493,7 +1634,7 @@ seat_real_start (Seat *seat)
             background_session = session;
             session = NULL;
         }
-      
+
         if (session)
         {
             DisplayServer *display_server;
@@ -1502,7 +1643,7 @@ seat_real_start (Seat *seat)
                 g_object_unref (seat->priv->session_to_activate);
             seat->priv->session_to_activate = g_object_ref (session);
 
-            display_server = create_display_server (seat, session_get_session_type (session));
+            display_server = create_display_server (seat, session);
             session_set_display_server (session, display_server);
             if (!display_server || !display_server_start (display_server))
             {
@@ -1533,20 +1674,7 @@ seat_real_start (Seat *seat)
         seat->priv->session_to_activate = g_object_ref (greeter_session);
         session = SESSION (greeter_session);
 
-        if (autologin_timeout)
-        {
-            gchar *value;
-
-            value = g_strdup_printf ("%d", autologin_timeout);
-            greeter_set_hint (greeter_session, "autologin-timeout", value);
-            g_free (value);
-            if (autologin_username)
-                greeter_set_hint (greeter_session, "autologin-user", autologin_username);
-            if (autologin_guest)
-                greeter_set_hint (greeter_session, "autologin-guest", "true");
-        }
-
-        display_server = create_display_server (seat, session_get_session_type (session));
+        display_server = create_display_server (seat, session);
         session_set_display_server (session, display_server);
         if (!display_server || !display_server_start (display_server))
         {
@@ -1570,13 +1698,12 @@ seat_real_start (Seat *seat)
     {
         DisplayServer *background_display_server;
 
-        background_display_server = create_display_server (seat, session_get_session_type (background_session));
+        background_display_server = create_display_server (seat, background_session);
         session_set_display_server (background_session, background_display_server);
         if (!display_server_start (background_display_server))
             l_warning (seat, "Failed to start display server for background session");
     }
 
-    seat->priv->started = TRUE;
     return TRUE;
 }
 
@@ -1663,12 +1790,13 @@ seat_finalize (GObject *object)
 
     self = SEAT (object);
 
+    g_free (self->priv->name);
     g_hash_table_unref (self->priv->properties);
     for (link = self->priv->display_servers; link; link = link->next)
     {
         DisplayServer *display_server = link->data;
         g_signal_handlers_disconnect_matched (display_server, G_SIGNAL_MATCH_DATA, 0, 0, NULL, NULL, self);
-    }  
+    }
     g_list_free_full (self->priv->display_servers, g_object_unref);
     for (link = self->priv->sessions; link; link = link->next)
     {
@@ -1691,7 +1819,6 @@ seat_class_init (SeatClass *klass)
 {
     GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
-    klass->get_start_local_sessions = seat_real_get_start_local_sessions;
     klass->setup = seat_real_setup;
     klass->start = seat_real_start;
     klass->create_greeter_session = seat_real_create_greeter_session;
@@ -1743,12 +1870,7 @@ seat_class_init (SeatClass *klass)
 static gint
 seat_real_logprefix (Logger *self, gchar *buf, gulong buflen)
 {
-    Seat *seat = SEAT (self);
-    const gchar *name = seat_get_string_property (seat, "seat-name");
-    if (name)
-        return g_snprintf (buf, buflen, "Seat %s: ", name);
-    else
-        return g_snprintf (buf, buflen, "Seat: ");
+    return g_snprintf (buf, buflen, "Seat %s: ", SEAT (self)->priv->name);
 }
 
 static void
